@@ -14,9 +14,11 @@ from copy import copy, deepcopy
 import logging
 import os.path
 
+from gi.repository import GLib
+
 from udiskie.common import Emitter, samefile
 from udiskie.compat import filter
-from udiskie.dbus import DBusProxy, DBusProperties, DBusException, DBusService
+from udiskie.dbus import DBusProxy, DBusException, DBusService
 
 __all__ = ['Sniffer', 'Daemon']
 
@@ -34,7 +36,7 @@ def object_kind(object_path):
 
 def filter_opt(opt):
     """Remove ``None`` values from a dictionary."""
-    return {k: v for k,v in opt.items() if v is not None}
+    return {k: GLib.Variant(*v) for k,v in opt.items() if v[1] is not None}
 
 
 Interface = dict(
@@ -114,15 +116,20 @@ class OfflineProxy(object):
     in many cases as it is immune to a number of race conditions.
     """
 
-    def __init__(self, proxy, data):
+    def __init__(self, proxy, interface, data):
         """
         Initialize wrapper.
 
-        :param DBusProxy proxy: for dynamic property/method lookup
+        :param DBusObject proxy: for dynamic property/method lookup
         :param dict data: for static property lookup
         """
         self.property = AttrDictView(data)
-        self.method = proxy.method
+        self._proxy = proxy
+        self._interface = interface
+
+    @property
+    def method(self):
+        return self._proxy.get_interface(self._interface).method
 
 
 class OfflineInterfaceService(object):
@@ -145,10 +152,10 @@ class OfflineInterfaceService(object):
 
     def __getattr__(self, key):
         """Return a wrapper for the requested interface."""
-        key = Interface[key]
+        interface = Interface[key]
         try:
-            return OfflineProxy(DBusProxy(self._proxy, key),
-                                self._data[key])
+            return OfflineProxy(self._proxy, interface,
+                                self._data[interface])
         except:
             return NullProxy(key, self._proxy.object_path)
 
@@ -166,16 +173,17 @@ class OnlineInterfaceService(object):
         """
         Store DBus proxy.
 
-        :param dbus.proxies.ProxyObject proxy: DBus object for online access
+        :param DBusObject object: DBus object for online access
         """
         self._proxy = proxy
-        self._check = DBusProxy(proxy, Interface['Properties']).method.GetAll
+        properties = proxy.get_interface(Interface['Properties'])
+        self._check = properties.method.GetAll
 
     def __getattr__(self, key):
         """Return a wrapper for the requested interface."""
         try:
-            self._check(Interface[key])
-            return DBusProxy(self._proxy, Interface[key])
+            self._check('(s)', Interface[key])
+            return self._proxy.get_interface(Interface[key])
         except DBusException:
             return NullProxy(key, self._proxy.object_path)
 
@@ -337,15 +345,21 @@ class Device(object):
     # Drive methods
     def eject(self, auth_no_user_interaction=None):
         """Eject media from the device."""
-        return self._assocdrive._I.Drive.method.Eject(filter_opt({
-            'auth.no_user_interaction': auth_no_user_interaction
-        }))
+        return self._assocdrive._I.Drive.method.Eject(
+            '(a{sv})',
+            filter_opt({
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
 
     def detach(self, auth_no_user_interaction=None):
         """Detach the device by e.g. powering down the physical port."""
-        return self._assocdrive._I.Drive.method.PowerOff(filter_opt({
-            'auth.no_user_interaction': auth_no_user_interaction
-        }))
+        return self._assocdrive._I.Drive.method.PowerOff(
+            '(a{sv})',
+            filter_opt({
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
 
     #----------------------------------------
     # Block
@@ -486,18 +500,24 @@ class Device(object):
               options=None,
               auth_no_user_interaction=None):
         """Mount filesystem."""
-        return self._I.Filesystem.method.Mount(filter_opt({
-            'fstype': fstype,
-            'options': ','.join(options or []),
-            'auth.no_user_interaction': auth_no_user_interaction
-        }))
+        return self._I.Filesystem.method.Mount(
+            '(a{sv})',
+            filter_opt({
+                'fstype': ('s', fstype),
+                'options': ('s', ','.join(options or [])),
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
 
     def unmount(self, force=None, auth_no_user_interaction=None):
         """Unmount filesystem."""
-        return self._I.Filesystem.method.Unmount(filter_opt({
-            'force': force,
-            'auth.no_user_interaction': auth_no_user_interaction
-        }))
+        return self._I.Filesystem.method.Unmount(
+            '(a{sv})',
+            filter_opt({
+                'force': ('b', force),
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
 
     #----------------------------------------
     # Encrypted
@@ -522,9 +542,13 @@ class Device(object):
     # Encrypted methods
     def unlock(self, password, auth_no_user_interaction=None):
         """Unlock Luks device."""
-        object_path = self._I.Encrypted.method.Unlock(password, filter_opt({
-            'auth.no_user_interaction': auth_no_user_interaction
-        }))
+        object_path = self._I.Encrypted.method.Unlock(
+            '(sa{sv})',
+            password,
+            filter_opt({
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
         # UDisks2 may not have processed the InterfacesAdded signal yet.
         # Therefore it is necessary to query the interface data directly
         # from the DBus service:
@@ -532,9 +556,12 @@ class Device(object):
 
     def lock(self, auth_no_user_interaction=None):
         """Lock Luks device."""
-        return self._I.Encrypted.method.Lock(filter_opt({
-            'auth.no_user_interaction': auth_no_user_interaction
-        }))
+        return self._I.Encrypted.method.Lock(
+            '(a{sv})',
+            filter_opt({
+                'auth.no_user_interaction': ('b', auth_no_user_interaction),
+            })
+        )
 
     #----------------------------------------
     # derived properties
@@ -623,7 +650,7 @@ class Sniffer(UDisks2):
         if not self._is_valid_object_path(object_path):
             return None
         return Device(self, object_path, OnlineInterfaceService(
-            self._proxy._bus.get_object(self.BusName, object_path)))
+            self._proxy.object.bus.get_object(object_path)))
 
     update = get
 
@@ -643,59 +670,39 @@ class Daemon(Emitter, UDisks2):
         - device_changed  / job_failed
     """
 
-    mainloop = True
-
     def __init__(self, proxy=None):
 
         """Initialize object and start listening to UDisks2 events."""
 
-        event_names = (tuple(stem + suffix
-                             for suffix in ('ed','ing')
-                             for stem in (
-                                 'device_add',
-                                 'device_remov',
-                                 'device_mount',
-                                 'device_unmount',
-                                 'media_add',
-                                 'media_remov',
-                                 'device_unlock',
-                                 'device_lock',
-                                 'device_chang', ))
-                       + ('object_added',
-                          'object_removed',
-                          'job_failed'))
+        event_names = ['device_added',
+                       'device_removed',
+                       'device_mounted',
+                       'device_unmounted',
+                       'media_added',
+                       'media_removed',
+                       'device_unlocked',
+                       'device_locked',
+                       'device_changed',
+                       'job_failed']
         super(Daemon, self).__init__(event_names)
 
-        self._proxy = proxy or self.connect_service()
+        self._proxy = proxy = proxy or self.connect_service()
         self._log = logging.getLogger(__name__)
         self._objects = {}
 
-        bus = self._proxy._bus
-        bus.add_signal_receiver(
-            self._interfaces_added,
-            signal_name='InterfacesAdded',
-            dbus_interface=Interface['ObjectManager'],
-            bus_name=self.BusName)
-        bus.add_signal_receiver(
-            self._interfaces_removed,
-            signal_name='InterfacesRemoved',
-            dbus_interface=Interface['ObjectManager'],
-            bus_name=self.BusName)
-        bus.add_signal_receiver(
-            self._properties_changed,
-            signal_name='PropertiesChanged',
-            dbus_interface=Interface['Properties'],
-            bus_name=self.BusName,
-            path_keyword='object_path')
-        bus.add_signal_receiver(
-            self._job_completed,
-            signal_name='Completed',
-            dbus_interface=Interface['Job'],
-            bus_name=self.BusName,
-            path_keyword='job_name')
-        self._sync()
+        proxy.connect('InterfacesAdded', self._interfaces_added)
+        proxy.connect('InterfacesRemoved', self._interfaces_removed)
 
-        self.connect('object_added', self._object_added)
+        bus = proxy.object.bus
+        bus.connect(Interface['Properties'],
+                    'PropertiesChanged',
+                    None,
+                    self._properties_changed)
+        bus.connect(Interface['Job'],
+                    'Completed',
+                    None,
+                    self._job_completed)
+        self._sync()
 
     def _sync(self):
         """Synchronize state."""
@@ -714,7 +721,7 @@ class Daemon(Emitter, UDisks2):
             if not interfaces_and_properties:
                 return None
         interface_service = OfflineInterfaceService(
-            self._proxy._bus.get_object(self.BusName, object_path),
+            self._proxy.object.bus.get_object(object_path),
             interfaces_and_properties)
         return Device(self, object_path, interface_service)
 
@@ -732,24 +739,41 @@ class Daemon(Emitter, UDisks2):
         added = object_path not in self._objects
         self._objects[object_path] = interfaces_and_properties
         if added:
-            self.trigger('object_added', object_path)
+            kind = object_kind(object_path)
+            if kind in ('device', 'drive'):
+                self.trigger('device_added', self[object_path])
 
-    def _object_added(self, object_path):
-        """Internal event handler."""
-        kind = object_kind(object_path)
-        if kind in ('device', 'drive'):
-            self.trigger('device_added', self[object_path])
-        elif kind == 'job':
-            self._job_changed(object_path, False)
+        if Interface['Block'] in interfaces_and_properties:
+            slave = self[object_path].luks_cleartext_slave
+            if slave:
+                if not self._has_job(slave.object_path, 'device_unlocked'):
+                    self.trigger('device_unlocked', slave)
 
     # remove objects / interfaces
     def _detect_toggle(self, property_name, old, new, add_name, del_name):
         old_valid = old and bool(getattr(old, property_name))
         new_valid = new and bool(getattr(new, property_name))
         if add_name and new_valid and not old_valid:
-            self.trigger(add_name, new)
+            if not self._has_job(old.object_path, add_name):
+                self.trigger(add_name, new)
         elif del_name and old_valid and not new_valid:
-            self.trigger(del_name, new)
+            if not self._has_job(old.object_path, del_name):
+                self.trigger(del_name, new)
+
+    def _has_job(self, device_path, event_name):
+        job_interface = Interface['Job']
+        for path, interfaces in self._objects.items():
+            try:
+                job = interfaces[job_interface]
+                job_objects = job['Objects']
+                job_operation = job['Operation']
+                job_action = self._action_mapping[job_operation]
+                job_event = self._event_mapping[job_action]
+                if event_name == job_event and device_path in job_objects:
+                    return True
+            except KeyError:
+                pass
+        return False
 
     def _interfaces_removed(self, object_path, interfaces):
         """Internal method."""
@@ -765,6 +789,12 @@ class Daemon(Emitter, UDisks2):
                 self.get(object_path, new_state),
                 None, 'media_removed')
 
+        if Interface['Block'] in interfaces:
+            slave = self.get(object_path, old_state).luks_cleartext_slave
+            if slave:
+                if not self._has_job(slave.object_path, 'device_locked'):
+                    self.trigger('device_locked', slave)
+
         if not self._objects[object_path]:
             del self._objects[object_path]
             if object_kind(object_path) in ('device', 'drive'):
@@ -774,10 +804,10 @@ class Daemon(Emitter, UDisks2):
 
     # change interface properties
     def _properties_changed(self,
+                            object_path,
                             interface_name,
                             changed_properties,
-                            invalidated_properties,
-                            object_path):
+                            invalidated_properties):
         """
         Internal method.
 
@@ -802,7 +832,10 @@ class Daemon(Emitter, UDisks2):
                 'is_mounted',
                 self.get(object_path, old_state),
                 self.get(object_path, new_state),
-                'device_mounted', None)
+                'device_mounted', 'device_unmounted')
+        # There is no PropertiesChanged for the crypto device when it is
+        # unlocked/locked in UDisks2. Instead, this is handled by the
+        # InterfaceAdded/Removed handlers.
 
     # jobs
     _action_mapping = {
@@ -814,36 +847,41 @@ class Daemon(Emitter, UDisks2):
         'eject-media': 'eject', }
 
     _event_mapping = {
-        'filesystem-unmount': 'device_unmount',
-        'encrypted-unlock': 'device_unlock',
-        'encrypted-lock': 'device_lock', }
+        'mount': 'device_mounted',
+        'unmount': 'device_unmounted',
+        'unlock': 'device_unlocked',
+        'lock': 'device_locked',
+        'eject': 'media_removed',
+        'detach': 'device_removed'}
 
-    def _job_changed(self, job_name, completed):
-        job = self._objects[job_name][Interface['Job']]
-        event_name = self._event_mapping.get(job['Operation'])
-        if not event_name:
-            return
-        suffix = 'ed' if completed else 'ing'
-        for object_path in job['Objects']:
-            device = self[object_path]
-            self.trigger(event_name + suffix, device)
+    _check_success = {
+        'mount': lambda dev: dev.is_mounted,
+        'unmount': lambda dev: not dev or not dev.is_mounted,
+        'unlock': lambda dev: dev.is_unlocked,
+        'lock': lambda dev: not dev or not dev.is_unlocked,
+        'detach': lambda dev: not dev,
+        'eject': lambda dev: not dev or not dev.has_media
+    }
 
-    def _job_failed(self, job_name, message):
-        job = self._objects[job_name][Interface['Job']]
-        action = self._event_mapping.get(job['Operation'])
-        if not action:
-            return
-        for object_path in job['Objects']:
-            device = self[object_path]
-            self.trigger('job_failed', device, action, message)
-
-    def _job_completed(self, success, message, job_name):
+    def _job_completed(self, job_name, success, message):
         """
         Internal method.
 
         Called when a job of a long running task completes.
         """
+        job = self._objects[job_name][Interface['Job']]
+        action = self._action_mapping.get(job['Operation'])
+        if not action:
+            return
+        # We only handle events, which are associated to exactly one object:
+        object_path, = job['Objects']
+        device = self[object_path]
         if success:
-            self._job_changed(job_name, True)
+            # It rarely happens, but sometimes UDisks posts the
+            # Job.Completed event before PropertiesChanged, so we have to
+            # check if the operation has been carried out yet:
+            if self._check_success[action](device):
+                event_name = self._event_mapping[action]
+                self.trigger(event_name, device)
         else:
-            self._job_failed(job_name, message)
+            self.trigger('job_failed', device, action, message)
